#
# classes for capricorn cpu (hp8x,75C)
#
# assembler, de-assembler, simulator ? (need system roms)
#

import string

class AsmError(Exception):      # Error class for Asm
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class cpu:
    def __init__(self, param=None):
        """
        create a cpu
        """
        self._regs= [0] * 64        # registres
        self._arp = -1              # ard
        self._drp = -1              # drp
        self._ext = 0               # extended register
        self._flags = 0             # flags
        self._dad = {}              # {symbol : address} global rom
        self._symbol = {}           # {address : symbol} global rom 
        self._com = {}              # {symbol : commentaire} global rom (for de-assembler)
        self._ldad = {}             # {symbol : address} local
        self._lsymbol = {}          # {address : symbol} local
        self._llsymbols = {}        # {address : symbol} local '.' used (for assembler)
        self._lldad = {}            # {symbol : llsymbols} local '.' symbols (for assembler)
        self._lcom = {}             # {symbol : comments} local
        self._lref = {}             # {address : [use]} cross ref  (JSB)
        self._lsrp = {}             # {symbol : ['arp ', 'drp '] } rp status for Sxxxx
        self._adrp = {}             # {symbol : [arp, drp] } arp & drp status back from symbol
        self._psymbol = {}          # {symbol : address} possible local 
        self._pdad = {}             # {address : symbol} possible local
        self._mem = [0] * 65536     # memory
        self._stat = [0] * 65536    # memory ststus 0: data, 1:code, 3 fin, 4 ptr data (for de-assembler)
        self._asc = string.ascii_letters + string.punctuation + string.digits + ' ' # ascii for labels ...
        self._errors = {
            1: 'UNDERFLOW',
            2: 'OVERFLOW',
            3: 'COT/CSC=INF',
            4: 'TAN/SEC=INF',
            5: '0^NEG',
            6: '0^0',
            7: 'NULL DATA',
            8: '/ZEO',
            9: 'NEG^NON-INT',
            10: 'SQR(-)',
            11: 'ARG OUT OF RANGE',
            12: 'LOG(0)',
            13: 'LOG(-)',
            15: 'SYSTEM',
            16: 'CONTINUE BEFORE RUN',
            17: 'FOR NESTING',
            18: 'GOSUB',
            19: 'MEM OVF',
            21: 'ROM MISSING',
            22: 'SECURED',
            23: 'SELF TEST',
            24: 'MODE',
            25: 'BAD BIN LOAD',
            26: 'STACK OVFL',
            27: '85 ROM IS IGNORED',
            30: 'OPTION BASE',
            31: 'CHAIN',
            32: 'COM MISMATCH',
            33: 'DATA TYPE',
            34: 'NO DATA',
            35: 'DIM EXIST VRBL',
            36: 'DIM ILLEGAL',
            37: 'DUP FN',
            38: 'NO FN END',
            39: 'FN MISSING',
            40: 'FN PARAM',
            41: 'FN=',
            42: 'RECURSIVE FN CALL',
            43: 'NUMERIC INPUT',
            44: 'TOO FEW INPUTS',
            45: 'TOO MANY INPUTS',
            46: 'NEXT MISSING',
            47: 'NO MATCHING FOR',
            48: 'END',
            49: 'NULL DATA',
            50: 'BIN PROG MISSING',
            51: 'RETURN W/O GOSUB',
            52: 'IMAGE',
            53: 'PRINT USING',
            54: 'TAB',
            55: 'SUBSCRIPT',
            56: 'STRING OVF',
            57: 'MISSING LINE',
            60: 'WRITE PROTECT',
            63: 'DUP NAME',
            66: 'FILE CLOSED',
            67: 'FILE NAME',
            68: 'FILE TYPE',
            69: 'RANDOM OVF',
            70: 'READ',
            71: 'EOF',
            72: 'RECORD',
            80: ') EXPECTED',
            81: 'BAD EXPRESSION',
            82: 'STRING EXPR',
            83: '"," MISSING',
            84: 'EXCESS CHARS',
            85: 'EXPR TOO BIG',
            86: 'ILLEGAL AFTER THEN',
            87: 'BAD DIM STATEMENT',
            88: 'BAD STATEMENT',
            89: 'INVALID PARAM',
            90: 'LINE>99999',
            91: 'MISSING PARAM',
            92: 'SYNTAX',
            110: 'I/O CARD',
            111: 'IOP',
            112: 'M.S. ROM',
            120: 'NO M.S. DEVICE',
            124: 'FILES',
            125: 'VOLUME',
            126: 'MSUS',
            127: 'READ VFY',
            128: 'FULL',
            129: 'MEDIUM',
            130: 'DISC',
            131: 'TIMEOUT'}
        
        if param == '85':           # load globals symbols for HP85
            files = ('globals85.txt',)
        if param == '86':           # load globals symbols for HP86
            files = ('globals86.txt', 'globals86-p.txt',)
        for file in files:
            self._symb(file)

    def mfill(self, value=0):       # fill memory with value
        self._mem = [value] * 65536
        self._stat = [0] * 65536        

    def _symb(self, file):          # read a file and get DAD symbols as global
        f=open(file, 'r')
        for line in f:
            token = string.split(line)
            if token[1][0] != '!':
                if len(token) > 2:
                    if token[2] == 'DAD':
                        self._dad[token[1]] = eval('0' + token[3])
                        self._symbol[eval('0' + token[3])] = token[1]
                        self._com[token[1]] = string.join(token[5:])
##                  if token[2] == 'EQU':
##                      self._symbol[token[1]] = eval('0' + token[3])
##                      self._dad[eval('0' + token[3])] = token[1]
        f.close()
    def arp(self, param=None):      # set or read arp register
        if param == None:
            return self._arp
        else:
            self._arp = param

    def drp(self, param=None):      # set or read drp register
        if param == None:
            return self._drp
        else:
            self._drp = param

    def bound(self, reg):           # compute the bound limit for M acces mode from reg
        if reg < 32:
            return 2 - (reg % 2)
        else:
            return 8 - (reg % 8)

    def ext(self, param=None):      # read or set E register
        if param == None:
            return self._ext
        else:
            self._ext = param

    def r(self, reg, val=None):     # read or set a register as byte
        if val == None:
            return self._regs[reg]
        else:
            self._regs[reg] = val % 256

    def r2(self, reg, val=None):    # read or set a register as word
        if val== None:
            return self._regs[reg] * 256 + self._regs[reg+1]
        else:
            self._regs[reg] = val % 256
            self._regs[reg+1] = (val / 256) % 256

    def rm(self, reg, val=None):    # read or set a slice of register until bound as list
        if val == None:
            return self._regs[reg:param+self.bound(reg)]
        else:
            self._regs[reg:reg+self.bound(reg)] = val

    def mem_b(self, ad, val=None):  # read or set memory as byte
        if val == None:
            return self._mem[ad]
        else:
            self._mem[ad] = val
    def mem_w(self, ad, val=None):  # read or set memory as word
        if val == None:
            return self._mem[ad] + self._mem[ad + 1] * 0x100
        else:
            self._mem[ad] = val % 0x100
            self._mem[ad + 1] = val / 0x100

    def _deass_op(self, pc, base):  # de-assemble the code in pc address
        def _byt(byte):                 # convert byte in base
            if base == 8:
                c = '00'+oct(byte)
                c = c[-3:]
            else:
                c = '%02X' % byte
            return c
        def _rp(reg):                   # convert reg in R{#,*,n} (n always in octal)
            if reg == -1:
                return 'R#'
            if reg == 1:
                return 'R*'
            else:
                if reg < 8:
                    return 'R' + oct(reg)[-1:]
                else:
                    return 'R' + oct(reg)[-2:]
        def _xp(reg):                   # convert reg in X{#,*,n} (n always in octal)
            if reg == -1:
                return 'X#'
            if reg == 1:
                return 'X*'
            else:
                if reg < 8:
                    return 'X' + oct(reg)[-1:]
                else:
                    return 'X' + oct(reg)[-2:]
        def popdrp():                   # pop a drp code if possible
            if len(pdrp) > 0:
                pdrp.pop()
        def poparp():                   # pop a arp code if possible
            if len(parp) > 0:
                parp.pop()
        def readh(j):                   # read a value for a string with base
            if base == 8:
                return eval('0' + h[j:j + 3])
            else:
                return eval('0x'+ h[j:j + 2])
        c = ''                          # text of the opcode
        h = ''                          # string for codes (in base)
        t = 0           # following bytes   J 1: offset 1 byte relative jump
                        #                   I 2: immediate adressing 2 bytes
                        #                   G 3: offset 2 bytes relative jump (ADM R4,=offset <=> GTO
                        #                   S 4: adress 2 bytes sous-routine
                        #                   D 5: adress 2 bytes direct & indirect
                        #                   X 6: address base for X mode
                        #                   G 7: adress 2 bytes LDM R4,=
                        #                   R    reference to local binary
        n = 0           # number or following bytes needed to complete the opcode
        st = 1          # code stream still valid ?
        parp = []       # stack for arp codes
        pdrp = []       # stack for drp codes
        while c == '':
            op = self.mem_b(pc)
            self._stat[pc] = 1
            pc += 1
            if op < 0x40:           # codes from 0x00 to 0x3F
                self.arp(op)
                parp.append(op)
                h += _byt(op) + ' '
            elif op < 0x80:         # codes from 0x40 to 0x7F
                self.drp(op - 0x40)
                pdrp.append(op)
                h += _byt(op) + ' '
            elif op < 0x94:         # codes from 0x80 to 0x93
                p = op - 0x80
                c = ('EL','ER','LL','LR','IC','DC','TC','NC','TS','CL')[p / 2]
                c += ('B','M')[p % 2]
                c += ' ' + _rp(self.drp())
                popdrp()
                h += _byt(op) + ' '
            elif op < 0x98:         # codes from 0x94 to 0x97
                p = op - 0x94
                c = ('OR','XR')[p / 2]
                c += ('B','M')[p % 2]
                c += ' ' + _rp(self.drp()) + ',' + _rp(self.arp())
                popdrp()
                poparp()
                h += _byt(op) + ' '
            elif op < 0xA0:         # codes from 0x98 to 0x9F
                p = op - 0x98
                c = ('BIN','BCD','SAD','DCE','ICE','CLE','RTN','PAD')[p]
                h += _byt(op) + ' '
                if p == 6:                  # RTN
                    st = 0                  # end of code stream
                if p == 7:                  # PAD
                    self.arp(-1)
                    self.drp(-1)
            elif op < 0xC0:         # codes from 0xA0 to 0xBF
                p = op - 0xA0
                c = ('LD','ST')[(p / 2) % 2]
                c += ('B','M')[p % 2]
                c += ('','D','','I','D','D','I','I')[p / 4]
                c += ' ' + _rp(self.drp()) + ','
                popdrp()
                if (((p / 2) % 2) == 0) and (self.drp() == 4) and ((p % 2) == 1):   # LDM? R4
                    st = 2
                    if ((p / 4) == 0) or ((p / 4) == 2):
                        t = 7
                if p < 0x08:
                    c += _rp(self.arp())
                    poparp()
                elif (p < 0x0C):
                    c += '='
                    if (p % 2) == 0:
                        n = 1
                    else:       # LDM R,=a STM R,=a
                        n = self.bound(self.drp())
                        if n == 2:
                            if t != 7:
                                t = 2
                elif p < 0x10:
                    c += _rp(self.arp())
                    poparp()
                    if (((p / 2) % 2) == 1) and (self.arp() == 4) and ((p % 2) == 1):   # STM? ,R4
                        st = 3
                elif p < 0x14:  # ???D R,=a
                    c += '='
                    n = 2
                    t = 5
                elif p < 0x18:  # ???D X,a
                    c += _xp(self.arp()) + ','
                    poparp()
                    n = 2
                    t = 6
                elif p < 0x1C:  # ???I R,=a
                    c += '='
                    n = 2
                    t = 5
                elif p < 0x20:  # ???I X,a
                    c += _xp(self.arp()) + ','
                    poparp()
                    n = 2
                    t = 5
                h += _byt(op) + ' '
            elif op < 0xE0:         # codes from 0xC0 to 0xDF
                p = op - 0xC0
                c = ('CM','AD','SB','AN')[(p / 2) % 4]
                c += ('B','M')[p % 2]
                if p > 0x0F:
                    c += 'D'
                if (p != 0x06) and (p != 0x0E):
                    c += ' ' + _rp(self.drp()) + ','
                    popdrp()
                    if ((p % 2) == 1) and (((p / 2) % 4) > 0) and (self.drp() == 4):    # {AD|SB|AN}M R4,
                        st = 2
                if p == 0x06:
                    c = 'JSB ' + _xp(self.arp()) + ','
                    poparp()
                    n = 2
                    t = 6
                elif p == 0x0E:
                    c = 'JSB ='
                    n = 2
                    t = 4
                elif (p == 0x16) or (p == 0x1E):
                    c = '???'
                elif p < 0x08:
                    c += _rp(self.arp())
                    poparp()
                elif p < 0x10:
                    c += '='
                    if (p % 2) == 0:
                        n = 1
                    else:
                        n = self.bound(self.drp())
                        if n == 2:
                            if (((p / 2) % 4) == 1) and (self.drp() == 4):       # ADM R4,
                                t = 3
                                st = 2
                            else:
                                t = 2
                elif p < 0x18:
                    c += '='
                    n = 2
                    t = 5
                elif p < 0x20:
                    c += _rp(self.arp())
                    poparp()
                h += _byt(op) + ' '
            elif op < 0xF0:         # codes from 0xE0 to 0xEF
                p = op - 0xE0
                c = ('PO','PU')[(p / 4) % 2]
                c += ('B','M')[p % 2]
                c += ('D','I')[p / 8]
                c += ' ' + _rp(self.drp()) + ','
                popdrp()
                if (((p / 4) % 2) == 0) and ((p % 2) == 1) and (self.drp() == 4):   # POM? R4,
                    st = 2
                c += ('+','-')[(p / 2) % 2]
                c += _rp(self.arp())
                poparp()
                h += _byt(op) + ' '
            else:                   # codes from 0xF0 to 0xFF
                p = op - 0xF0
                c = 'J'
                c += ('MP','NO','OD','EV','NG','PS','NZ','ZR','EN','EZ','NC','CY','LZ','LN','RZ','RN')[p]
                c += ' '
                h += _byt(op) + ' '
                n = 1
                t = 1
                if p == 0:
                    st = 0
        # ARP and DRP codes not used
        if base == 8:
            s = 4
        else:
            s = 3
        i = 0
        while i < len(parp): # remove ARP not used
            j = 0
            while readh(j) != parp[i]:
                j += s
            h = h[:j] + h[j + s:]
            i += 1
        i = 0
        while i < len(pdrp): # remove DRP not used
            j = 0
            while readh(j) != pdrp[i]:
                j += s
            h = h[:j] + h[j + s:]
            i += 1
        # new adress, opcode, string, number of bytes to follow, type of bytes, valid stream, stack arp, stack drp
        return pc, c, h, n, t, st, parp, pdrp
    def deass_bin(self, code, csym={}, dsym={}, lcom={}, stat={}, adrp= {}, level=0, disp=0, base=8, rom=0, out=None):
        # code  : a list of bytes []
        # csym  : supplementary {symbol, address} of streams of code (human created)
        # dsym  : supplementary {symbol, address} of local symbols for data & reference only (human craeted)
        # lcom  : comments for local symbols {symbol, comment} (human created)
        # stat  : special memory status hint to help data presentation (human created)
        # adrp  : supplementary ARP & DRP status for symbols
        # level : the # of the pass : 0, 1 or 2
        # disp  : diagnostic display : 1
        # base  : in which base should we create the de-assembly
        # rom   : if the code is a ROM dump : 1
        # out   : the name of a text file to be created or None
        #
        # supplementary informations are not always used in first pass
        # a binary is always relocated at 0x0000, a ROM stays at 0x6000
        # first pass search streams of code and define associated labels (with help of csym)
        # second pass use all supplementaries and should find ALL code
        # third pass can help to define precisely data blocks
        def _byt(byte):                     # convert byte in base
            if base == 8:
                c = '00'+oct(byte)
                c = c[-3:]
            else:
                c = '%02X' % byte
            return c
        def _adr(ad):                       # convert adress in base
            if base == 8:
                return string.zfill(oct(ad),6)
            else:
                return '%04X' % ad
        def _out(str):                      # write a string or print it
            if out == None:
                print str
            else:
                out.write(str + '\n')
        def _emit(ad, h, c, disp, r=''):    # shape output of code (take care of local symbol as label)
            if r != '':
                r = '! '+r
            if disp == 1:
                first = 1
                while (h != '') or (first == 1):
                    if ad in self._lsymbol:
                        lab = self._lsymbol[ad]
                        if lab in self._lsrp:
                            if r != '':
                                r += ' (D:'
                            else:
                                r = '! (D:'
                            k = self._lsrp[lab][1]
                            k.sort()
                            for i in range(len(k)):
                                if k[i] >= 0:
                                    if k[i] < 8:
                                        r += oct(k[i])[-1:]
                                    else:
                                        r += oct(k[i])[-2:]
                                else:
                                    r += '?'
                                if i < len(k) - 1:
                                    r += ','
                            r += '/A:'
                            k = self._lsrp[lab][0]
                            k.sort()
                            for i in range(len(k)):
                                if k[i] >= 0:
                                    if k[i] < 8:
                                        r += oct(k[i])[-1:]
                                    else:
                                        r += oct(k[i])[-2:]
                                else:
                                    r += '?'
                                if i < len(k) - 1:
                                    r += ','
                            r += ')'
                    else:
                        lab = ''
                    if first == 1:
                        if base == 8:
                            _out(_adr(ad) + ' ' + h[:8] + string.ljust('', 8-len(h[:8])) + string.ljust(lab,9) + string.ljust(c, 24) + r)
                        else:
                            _out(_adr(ad) + ' ' + h[:6] + string.ljust('', 6-len(h[:6])) + string.ljust(lab,9) + string.ljust(c, 24) + r)                            
                        first = 0
                    else:
                        if base == 8:
                            _out(_adr(ad) + ' ' + h[:8] + string.ljust('', 8-len(h[:8])) + string.ljust(lab,9))
                        else:
                            _out(_adr(ad) + ' ' + h[:6] + string.ljust('', 6-len(h[:6])) + string.ljust(lab,9))
                    if base == 8:
                        h = h[8:]
                    else:
                        h = h[6:]
                    ad += 2
        def _addsym(name, address):         # add a local symbol,address if possible
            if address not in self._lsymbol:
                self._lsymbol[address] = name
                self._ldad[name] = address
        def _addpsym(name, address):        # add a potential local symbol if possible
            if address not in self._psymbol:
                self._psymbol[address] = name
                self._pdad[name] = address
        def _nsym(ad):                      # find the address of the next symbol (pass > 0)
            ad += 1
            while (ad not in self._lsymbol) and (ad < end):
                ad += 1
            return ad
        def _asc(ad):                       # find the longest ascii string until next symbol
            fi = _nsym(ad)
            while (chr(self.mem_b(ad)) in self._asc) and (self._stat[ad] == 0) and (ad < fi):
                ad += 1
            return ad
        def _asp(ad):                       # find the longest ascii (+0x80) string until next symbol
            de = ad
            fi = _nsym(ad)
            while (chr(self.mem_b(ad)) in self._asc) and (self._stat[ad] == 0) and (ad < fi):
                ad += 1
            if (ad > de) and (self.mem_b(ad) >= 128):
                if (chr(self.mem_b(ad) - 128) in self._asc) and (self._stat[ad] == 0):
                    return ad + 1
            return de
        def _str(de, fi):                   # return the string betwen de & fi 
            s = ''
            while de < fi:
                s += chr(self.mem_b(de) % 128)
                de += 1
            return s
        def _bsz(ad):                       # find the longest block of zeros
            fi = _nsym(ad)
            while (self.mem_b(ad) == 0x0) and (self._stat[ad] == 0) and (ad < fi):
                ad += 1
            return ad
        def _ncode(ad):                     # find address of following code
            fi = _nsym(ad)
            while (self._stat[ad] == 0) and (ad < fi):
                ad += 1
            return ad
        def _rco(ad, typ = 1):              # change status of memory at ad
            if self._stat[ad] == 0:
                self._stat[ad] = typ
            return self.mem_b(pc), pc + 1
        def _entete_bin(code):              # binary header & relocation at 0x000            
            pc = org
# org
            c = 'ORG ' + _adr(org)
            h = ''
            _emit (org + 0x0000, h, c, disp)
# org
            org_0 = self.mem_w(0x18)        # origine du binaire original
            end_0 = org_0 + len(code)       # fin du binaire original
            end = org + len(code)           # fin du binaire reloge
            self._stat[end] = 3             # status fin binaire
# nam
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x18, self.mem_w(org + 0x18) - org_0 + org)    # org_0
            c = 'NAM ' + _byt(self.mem_b(org + 0x0007)) + ','
            h = ''
            for i in range(4):
                if chr(self.mem_b(org + i)) in self._asc:
                    c += chr(self.mem_b(org + i))
            i = 0
            while i < 26:
                h += _byt(self.mem_b(org + i)) + ' '
                i += 1
            
            _emit(org + 0x0000, h, c, disp)
            
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x1A, self.mem_w(org + 0x1A) - org_0 + org)    # RUNTIME
                self.mem_w(org + 0x1C, self.mem_w(org + 0x1C) - org_0 + org)    # ASCIIS
                self.mem_w(org + 0x1E, self.mem_w(org + 0x1E) - org_0 + org)    # PARSE
                self.mem_w(org + 0x20, self.mem_w(org + 0x20) - org_0 + org)    # ERRMSG
                self.mem_w(org + 0x22, self.mem_w(org + 0x22) - org_0 + org)    # INIT
# runtime
            nt = 1
            ad = self.mem_w(org + 0x1A)
            _addsym('RUNTIM', ad)
            c = 'DEF RUNTIM'
            h = _byt(self.mem_b(org + 0x1A)) + ' ' + _byt(self.mem_b(org + 0x1B)) + ' '
            _emit(org + 0x001A, h, c, disp)
            if level == 0:
                if self.mem_w(ad) != 0xFFFF:
                        self.mem_w(ad, self.mem_w(ad) - org_0 + org)                    
                ad += 2
                while ad + 1 < self.mem_w(org + 0x1E) :
                    if self.mem_w(ad) != 0xFFFF:
                        self.mem_w(ad, self.mem_w(ad) - org_0 + org)
                        if self.mem_w(ad) != 0x0000:
                            self._adr.append(self.mem_w(ad))
                            _addsym('RUN%03d' % nt, self.mem_w(ad))
                            if disp:
                                print 'RUN%03d %04X %04X' % (nt, self.mem_w(ad), self.mem_w(ad) - org + org_0)
                    nt += 1
                    ad += 2
# asciis
            ad = self.mem_w(org + 0x1C)
            _addsym('ASCIIS', ad)
            c = 'DEF ASCIIS'
            h = _byt(self.mem_b(org + 0x1C)) + ' ' + _byt(self.mem_b(org + 0x1D)) + ' '
            _emit(org + 0x001C, h, c, disp)
# parse
            nt = 1
            ad = self.mem_w(org + 0x1E)
            _addsym('PARSE', ad)
            c = 'DEF PARSE'
            h = _byt(self.mem_b(org + 0x1E)) + ' ' + _byt(self.mem_b(org + 0x1F)) + ' '
            _emit(org + 0x001E, h, c, disp)
            if level == 0:
                if self.mem_w(ad) != 0xFFFF:
                        self.mem_w(ad, self.mem_w(ad) - org_0 + org)                    
                ad += 2
                while ad + 1 < self.mem_w(org + 0x1C):
                    if self.mem_w(ad) != 0xFFFF:
                        self.mem_w(ad, self.mem_w(ad) - org_0 + org)
                        if self.mem_w(ad) != 0x0000:
                            self._adr.append(self.mem_w(ad))
                            _addsym('PAR%03d' % nt, self.mem_w(ad))
                            if disp:
                                print 'PAR%03d %04X %04X' % (nt, self.mem_w(ad), self.mem_w(ad) - org + org_0)
                    nt += 1
                    ad += 2
# ermsg
            ad = self.mem_w(org + 0x20)
            _addsym('ERMSG', ad)
            c = 'DEF ERMSG'
            h = _byt(self.mem_b(org + 0x20)) + ' ' + _byt(self.mem_b(org + 0x21)) + ' '
            _emit(org + 0x0020, h, c, disp)
# init
            ad = self.mem_w(org + 0x22)
            _addsym('INIT', ad)
            self._adr.append(ad)
            c = 'DEF INIT'
            h = _byt(self.mem_b(org + 0x22)) + ' ' + _byt(self.mem_b(org + 0x23)) + ' '
            _emit(org + 0x0022, h, c, disp)
            return org, end, org_0, end_0


        def _entete_rom(code):              # ROM header
            pc = org
# org
            c = 'ORG ' + _adr(org)
            h = ''
            _emit (org + 0x0000, h, c, disp)
# org
            org_0 = 0x6000                  # origine de la rom originale
            end_0 = org_0 + len(code)       # fin de la rom originale
            end = org + len(code)           # fin de la rom relogee

            self._stat[end] = 3             # status fin rom
# code 
            c = 'BYT ' + _byt(self.mem_b(org + 0x00)) + ',' + _byt(self.mem_b(org + 0x01))
            h = _byt(self.mem_b(org + 0x00)) + ' ' + _byt(self.mem_b(org + 0x01)) + ' '
            _emit(org + 0x0000, h, c, disp)
# runtime
            nt = 1
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x02, self.mem_w(org + 0x02) - org_0 + org)
            ad = self.mem_w(org + 0x02)
            _addsym('RUNTIM', ad)
            c = 'DEF RUNTIM'
            h = _byt(self.mem_b(org + 0x02)) + ' ' + _byt(self.mem_b(org + 0x03)) + ' '
            _emit(org + 0x0002, h, c, disp)
            if level == 0:
                while ad <= self.mem_w(org + 0x06):
                    self.mem_w(ad, self.mem_w(ad) - org_0 + org)
                    if self.mem_w(ad) != 0x0000:
                        self._adr.append(self.mem_w(ad))
                        if nt == 1:
                            _addsym('INIT', self.mem_w(ad))
                        else:
                            _addsym('RUN%03d' % (nt - 1), self.mem_w(ad))
                    nt += 1
                    ad += 2
# asciis
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x04, self.mem_w(org + 0x04) - org_0 + org)
            ad = self.mem_w(org + 0x04)
            _addsym('ASCIIS', ad)
            c = 'DEF ASCIIS'
            h = _byt(self.mem_b(org + 0x04)) + ' ' + _byt(self.mem_b(org + 0x05)) + ' '
            _emit(org + 0x0004, h, c, disp)
# parse
            nt = 1
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x06, self.mem_w(org + 0x06) - org_0 + org)
            ad = self.mem_w(org + 0x06)
            _addsym('PARSE', ad)
            c = 'DEF PARSE'
            h = _byt(self.mem_b(org + 0x06)) + ' ' + _byt(self.mem_b(org + 0x07)) + ' '
            _emit(org + 0x0006, h, c, disp)
            ad += 2
            if level == 0:
                while ad < self.mem_w(org + 0x04):
                    self.mem_w(ad, self.mem_w(ad) - org_0 + org)
                    if self.mem_w(ad) != 0x000:
                        self._adr.append(self.mem_w(ad))
                        _addsym('PAR%03d' % nt, self.mem_w(ad))
                    nt += 1
                    ad += 2
# ermsg
            if level == 0:      # relogement passe 0
                self.mem_w(org + 0x08, self.mem_w(org + 0x08) - org_0 + org)
            ad = self.mem_w(org + 0x08)
            if level == 0:
                self._adr.append(ad)
            _addsym('ERMSG', ad)
            c = 'DEF ERMSG'
            h = _byt(self.mem_b(org + 0x08)) + ' ' + _byt(self.mem_b(org + 0x09)) + ' '
            _emit(org + 0x0008, h, c, disp)
# init
##            if level == 0:      # relogement passe 0
##                self.mem_w(org + 0x0A, self.mem_w(org + 0x0A) - org_0 + org)
##            ad = self.mem_w(org + 0x0A)
##            _addsym('INIT', ad)
##            self._adr.append(ad)
##            c = 'DEF INIT'
##            h = _byt(self.mem_b(org + 0x0A)) + ' ' + _byt(self.mem_b(org + 0x0B)) + ' '
##            _emit(org + 0x000A, h, c, disp)
# test init
##            if level == 0:      # relogement passe 0
##                self.mem_w(org + 0x0C, self.mem_w(org + 0x0C) - org_0 + org)
##            ad = self.mem_w(org + 0x0C)
##            _addsym('TESTINIT', ad)
##            self._adr.append(ad)
##            c = 'DEF TESTINIT'
##            h = _byt(self.mem_b(org + 0x0C)) + ' ' + _byt(self.mem_b(org + 0x0D)) + ' '
##            _emit(org + 0x000C, h, c, disp)

# reloge base parse, runtime
##            if level == 0:      # relogement passe 0
##                ad = self.mem_w(org + 0x02)
##                self.mem_w(ad, self.mem_w(ad) - org_0 + org)
##                ad = self.mem_w(org + 0x06)
##                self.mem_w(ad, self.mem_w(ad) - org_0 + org)
            return org, end, org_0, end_0

        self._psymbol = {}  # possible local symbols
        self._pdad = {}     # possible local adresses
        self._lref = {}     # cross references  (JSB)
        if rom == 1:        # for a ROM
            org = 0x6000
            end = 0x0000
            org_0 = 0x6000
            end_0 = 0x0000
        else:               # for a binary
            org = 0x0000
            end = 0x0000
            org_0 = 0x0000
            end_0 = 0x0000
        
        if level == 0:      # initialisation for first pass
            self.mfill(0x00)    # zero's memory
            self._lsymbol = {}  # local symbols
            self._ldad = {}     # local adresses
            self._adr = []      # [adresses] of stream of code
            for i in csym:      # add supplementary symbols & address of streams of code
                _addsym(i, csym[i])
                self._adr.append(self._ldad[i])
            for i in lcom:      # add supplementary comments for symbols
                self._lcom[i] = lcom[i]
            pc = org
            for op in code:     # load code in memory
                self._mem[pc] = op
                pc += 1
            for i in stat:      # load given status of memory for addresses
                j = i
                for k in stat[i]:
                    self._stat[j] = k
                    j += 1
            for i in adrp:      # load given state of ARP & DRP for symbols 
                self._adrp[i] = [adrp[i][0], adrp[i][1]]
        for i in dsym:      # add supplementary data symbols 
            _addsym(i, dsym[i])
        if rom == 1:        # header for ROM (streams of code, ....)
            org, end, org_0, end_0 = _entete_rom(code)
        else:               # header for binary (streams of code, ...)
            org, end, org_0, end_0 = _entete_bin(code)
        if level >= 1:      # initialisation for next passes
            if rom == 1:        # for a ROM : first code address
                self._adr = [org + 0x000A]
                pc = org + 0x000A
            else:               # for a binary : fisrt code address
                self._adr = [org + 0x0024]
                pc = org + 0x0024
        else:               # for first pass : pop an address from the list of streams
            pc = self._adr.pop(0)

        self.arp(-1)        # unknown ARP
        self.drp(-1)        # unknown DRP
        while 1:            # GO
            if (level == 0) or (self._stat[pc] == 1):   # code (status or first pass) 
                de = pc                                 # store start address
                r = ''
                pc, c, h, n, t, st, parp, pdrp = self._deass_op(pc, base)   # opcode
                while (len(parp) > 0) or (len(pdrp) > 0):   # unused ARP et DRP codes
                    op = self.mem_b(de)
                    if op < 0x40:                           # ARP
                        if op == parp[0]:
                            _emit(de, _byt(op), 'ARP ' + oct(op)[-2:], disp, 'ARP')
                            parp.pop(0)
                            de += 1
                        else:
                            raise AsmError, 'PB ARP'
                    elif op < 0x80:                         # DRP
                        if op == pdrp[0]:
                            _emit(de, _byt(op), 'DRP ' + oct(op-0x40)[-2:], disp, 'DRP')
                            pdrp.pop(0)
                            de += 1
                        else:
                            raise AsmError, 'PB DRP'
                    else:                                   # should not happen
                        raise AsmError, 'PB ?RP'
                if t == 1:                      # J 1 : relative jump 1 byte
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    if op > 0x80:
                        ad = pc - (0x100-op)
                    else:
                        ad = pc + op
                    if level == 0:
                        self._adr.append(ad)        # new stream of code
                    _addsym('J' + _adr(ad), ad)     # add a new J(ump) symbol
                    c += self._lsymbol[ad]
                elif t == 2:                    # I 2 : 2 bytes as immediate mode
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if ad in self._lsymbol:                 # if a symbol exists
                        c += self._lsymbol[ad]              # use it as text
                    else:
                        if (ad >= org_0) and (ad < end_0):  # if in the binary
                            r += 'I' + _adr(ad)
                            _addpsym(r, ad)                 # add a new I(mmediate) possible symbol
                            c += r
                        elif ad in self._symbol:            # existing global symbol
                            c += self._symbol[ad]
                            r = '> ' + self._com[self._symbol[ad]]  # comment
                        else:
                            c += _adr(ad)
                elif t == 3:                    # G 3 : relative jump 2 bytes (ADM R4,= )
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if op > 0x8000:
                        ad = pc - (0x10000-op)
                    else:
                        ad = pc + op
                    if level == 0:                  # first pass
                        if (ad >= org_0) and (ad < end_0):  # if in the binary
                            self._adr.append(ad)            # new stream of code
                            _addsym('G' + _adr(ad), ad)     # add a new G(oto) symbol
                            c += self._lsymbol[ad]
                    else:                           # find an existing symbol ?
                        if ad in self._lsymbol:         # a local one
                            c += self._lsymbol[ad]
                        elif ad in self._symbol:        # a global one
                            c += self._symbol[ad]
                            r = '> ' + self._com[self._symbol[ad]]
                        else:                           # none
                            c += _adr(ad)
                            r = 'UNKNOWN GTO'
                elif t == 4:                    # S 4 : sub-routine address
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    rp = 0                      # ARP & DRP unknown when returning
                    if rom == 1:                    # for ROM
                        if ad in self._lsymbol:                 # local symbol existing ?
                            c += self._lsymbol[ad]
                            if ad in self._lref:                # cross references
                                self._lref[ad].append(_adr(de))
                            else:
                                self._lref[ad] = [_adr(de)]
                            if self._lsymbol[ad] not in self._lsrp:    # ARP & DRP
                                self._lsrp[self._lsymbol[ad]] = [[self.arp()], [self.drp()]]
                            else:
                                if self.arp() not in self._lsrp[self._lsymbol[ad]][0]:
                                    self._lsrp[self._lsymbol[ad]][0].append(self.arp())
                                if self.drp() not in self._lsrp[self._lsymbol[ad]][1]:
                                    self._lsrp[self._lsymbol[ad]][1].append(self.drp())
                            if self._lsymbol[ad] in self._lcom:         # local comment
                                r += '> ' + self._lcom[self._lsymbol[ad]]
                            if self._lsymbol[ad] in self._adrp:         # ARP & DRP value when returning
                                r += ' (D'
                                a = self._adrp[self._lsymbol[ad]][1]
                                if a < 8:
                                    r += oct(a)[-1:]
                                else:
                                    r += oct(a)[-2:]
                                r += '/A'
                                a = self._adrp[self._lsymbol[ad]][0]
                                if a < 8:
                                    r += oct(a)[-1:]
                                else:
                                    r += oct(a)[-2:]
                                r += ') <'
                        elif ad in self._symbol:                # global symbol existing ?
                            c += self._symbol[ad]
                            r = '> ' + self._com[self._symbol[ad]]
                            if (ad >= 0x6000) and (ad < 0x8000):
                                r += ' (EXTERNAL ROM)'
                        elif (ad >= org_0) and (ad < end_0):    # inside the ROM
                            if level == 0:                      # first pass ?
                                self._adr.append(ad)                # add new stream of code
                            _addsym('S' + _adr(ad), ad)     # add a new S(ubroutine) symbol
                            if ad in self._lref:            # cross references
                                self._lref[ad].append(_adr(de))
                            else:
                                self._lref[ad] = [_adr(de)]
                            if self._lsymbol[ad] not in self._lsrp:    # ARP & DRP
                                self._lsrp[self._lsymbol[ad]] = [[self.arp()], [self.drp()]]
                            else:
                                if self.arp() not in self._lsrp[self._lsymbol[ad]][0]:
                                    self._lsrp[self._lsymbol[ad]][0].append(self.arp())
                                if self.drp() not in self._lsrp[self._lsymbol[ad]][1]:
                                    self._lsrp[self._lsymbol[ad]][1].append(self.drp())
                            c += self._lsymbol[ad]
                        else:                                   # unknown sub-routine
                            c += _adr(ad)
                            r = 'UNKNOWN SP'
                    else:                           # for binary
                        if ad in self._symbol:                  # global symbol existing ?
                            c += self._symbol[ad]
                            r = '> ' + self._com[self._symbol[ad]]
                            if (ad >= 0x6000) and (ad < 0x8000):
                                r += ' (EXTERNAL ROM)'
                        elif ad in self._lsymbol:               # local symbol existing ?
                            c += self._lsymbol[ad]
                        elif (ad >= org_0) and (ad < end_0):    # inside binary ?
                            if level == 0:                          # first pass
                                self._adr.append(ad)                    # add a new stream of code
                            _addsym('S' + _adr(ad), ad)             # add a new S(ub-routine) local symbol
                            c += self._lsymbol[ad]
                        else:                                   # unknown sub-routine
                            c += _adr(ad)
                            r = 'UNKNOWN SP'
                    if ad in self._symbol:                      # for known global symbol
                        if self._symbol[ad] == 'ROMJSB':            # if ROMJSB define the next 3 bytes
                            _emit(de, h, c, disp, r)                    # emit the current opcode
                            de = pc
                            r = ''
                            c = 'DEF '                                  # DEF the next 2 bytes
                            if self.mem_w(pc) in self._symbol:          # as symbol
                                c += self._symbol[self.mem_w(pc)]
                                r = self._com[self._symbol[self.mem_w(pc)]]
                            else:                                       # as address
                                c += _adr(self.mem_w(pc))
                            h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                            pc += 2
                            _emit(de, h, c, disp, r)                    # emit the DEF statement
                            de = pc
                            r = ''
                            c = 'VAL '                                  # VAL the following byte
                            if self.mem_b(pc) == 0xD0:                  # as MSROM#
                                c += 'MSROM#'
                            elif self.mem_b(pc) == 0x28:                # as ASMROM#
                                c += 'ASMROM#'
                            else:
                                c+= _byt(self.mem_b(pc))
                            h = _byt(self.mem_b(pc)) + ' '              # as byte
                            pc += 1
                        if (self._symbol[ad] == 'ERROR+') or (self._symbol[ad] == 'ERROR'): # if ERROR(+)
                            _emit(de, h, c, disp, r)                    # emit the current opcode
                            de = pc
                            r = ''
                            c = 'BYT '+ _byt(self.mem_b(pc))            # BYT the next byte
                            h = _byt(self.mem_b(pc)) + ' '
                            if self.mem_b(pc) in self._errors:          # if error defined
                                r = 'ERROR : ' + self._errors[self.mem_b(pc)]   # use text as comment
                            pc += 1
                            if self._symbol[ad] == 'ERROR+':            # ERROR+ end the stream of code
                                st = 0
                    if rp == 0:                                 # assert the ARP & DRP as unknown (always actually)
                        self.arp(-1)
                        self.drp(-1)
                elif t == 5:                    # D 5 : direct and indirect mode adress
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc =_rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if ad in self._lsymbol:                 # is local sumbol existing ?
                        c += self._lsymbol[ad]
                    elif ad in self._symbol:                # is global symbol existing ?
                        c += self._symbol[ad]
                        r = '> ' + self._com[self._symbol[ad]]  # comment
                    elif (ad >= org_0) and (ad < end_0):    # inside binary ?
                        r = 'D' + _adr(ad)
                        _addsym(r, ad)                          # add a new D(irect) local symbol
                        c += r
                    else:                                   # as address
                        c += _adr(ad)
                elif t == 6:                    # X 6 : X adressing mode base
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc =_rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if ad in self._lsymbol:                 # is local symbol existing ?
                        c += self._lsymbol[ad]
                    elif ad in self._symbol:                # is global symbol existing ?
                        c += self._symbol[ad]
                        r = 'GLOBAL ?'
                    elif (ad >= org_0) and (ad < end_0):    # inside binary ?
                        r = 'X' + _adr(ad)
                        _addsym(r, ad)                          # add a new X() local symbol
                        c += r
                    else:                                   # as address
                        c += _adr(ad)
                elif t == 7:                    # G 7 : absolute jump (LDM R4,= )
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if level == 0:              # first pass
                        if (ad >= org_0) and (ad < end_0):  # inside binary ? 
                            self._adr.append(ad)                # add a new stream of code
                            _addsym('G' + _adr(ad), ad)         # add a new G(oto) local symbol
                            c += self._lsymbol[ad]
                    else:                       # further passes
                        if ad in self._lsymbol:         # is a local symbol existing ?
                            c += self._lsymbol[ad]
                        elif ad in self._symbol:        # is a global symbol existing ?
                            c += self._symbol[ad]
                            r = '> ' + self._com[self._symbol[ad]]
                        else:                           # as address
                            c += _adr(ad)
                            r = 'UNKNOWN GTO'
                elif n == 1:                    # 1 usual byte
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    c += _byt(op)
                    if chr(op) in self._asc:        # if ascii ?
                        r = "'" + chr(op) + "'"         # add representation in comment
                elif n == 2:                    # 2 usual bytes
                    ad, pc = _rco(pc)
                    h += _byt(ad) + ' '
                    op, pc = _rco(pc)
                    h += _byt(op) + ' '
                    ad += op * 0x100
                    if ad in self._lsymbol:                 # is a local symbol existing ?
                        c += self._lsymbol[ad]
                    elif (ad >= org_0) and (ad < end_0):    # is a global symbol existing ?
                        r = 'I' + _adr(ad)
                        _addpsym(r, ad)
                        c += r
                    else:                                   # as address
                        c += _adr(ad)
                elif n > 2:                     # more than 2 bytes (can use memory status)
                    s = ''
                    l = n
                    while n > 0:
                        if self._stat[pc] == 11:    # special case : 1 byte
                            n -= 1
                            op, pc = _rco(pc)
                            h += _byt(op) + ' '
                            if c == '':                 # no current opcode ?
                                c += 'BYT '                 # BYT opcode if needed
                            c += _byt(op)
                            _emit(de, h, c,disp, r)     # emit the opcode
                            c = ''
                            h = ''
                            r = ''
                            de = pc                     # initialisation for the next opcode
                        elif self._stat[pc] == 12:  # special case : 2 bytes as address
                            n -= 2
                            c += 'DEF '                 # DEF opcode
                            ad, pc = _rco(pc)
                            h += _byt(ad)+ ' '
                            op, pc = _rco(pc)
                            h += _byt(op) + ' '
                            ad += op * 0x100
                            if ad in self._lsymbol:     # is a local symbol existing ?
                                c += self._lsymbol[ad]
                            else:                       # as address
                                c += _adr(ad)
                            _emit(de, h, c, disp, r)    # emit the opcode
                            c = ''
                            h = ''
                            r = ''
                            de = pc                     # initialise for the next opcode
                        else:                       # nothing special : a bunch of BYT
                            n -= 1
                            op, pc = _rco(pc)
                            h += _byt(op) + ' '
                            if c == '':                 # no current opcode 
                                c = 'BYT '                  # BYT opcode if needed
                            c += _byt(op)
                            if chr(op) in self._asc:
                                s += chr(op)
                            if n > 0:
                                c += ','
                    if len(s) == l:
                        r += "'" + s + "'"
                if st > 1:                      # direct change of R4
                    r += ' (PC CHANGED)'            # comment that
                if c != '':                     # if some opcode 
                    _emit(de, h, c, disp, r)        # emit it
                if st != 1:                     # stream of code finished ?
                    self.arp(-1)
                    self.drp(-1)
                    if level == 0:                  # first pass ?
                        if self._adr == []:             # no more streams
                            break                           # first pass finished
                        else:                           # find a new stream not already treated
                            pc = self._adr.pop(0)
                            while (self._stat[pc] == 1) and (self._adr != []):
                                pc = self._adr.pop(0)
                            if self._stat[pc] == 1:         # no more streams
                                break                           # first pass finished
                        if (pc < 0x0000) or (pc > 0xFFFF):
                            raise AsmError, 'ADRESS ???'
                if (level >= 1):                # further passes
                    if pc == end:                   # finished ?
                        break
                    elif pc > end:                  # some consistency check
                        raise AsmError, 'AFTER END OF CODE'
            if (level >= 1) and (self._stat[pc] != 1):  # data stuff treated only after first pass
                # _stat      0 : data
                #            1 : code
                #            2 :
                #            3 : end of code
                if self._ldad['RUNTIM'] == pc:  # header building for RUNTIM
                    if rom == 1:                    # for ROM
                        fi = self.mem_w(org + 0x06)
                    else:                           # for binary
                        fi = self.mem_w(org + 0x1E)
                    while pc < fi:
                        c = 'DEF '
                        if self.mem_w(pc) in self._lsymbol:
                            c += self._lsymbol[self.mem_w(pc)]
                        elif self.mem_w(pc) in self._symbol:
                            c += self._symbol[self.mem_w(pc)]
                        else:
                            c += _adr(self.mem_w(pc))
                        h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                        _emit(pc, h, c, disp)
                        pc += 2
                elif self._ldad['PARSE'] == pc: # header building for PARSE
                    if rom == 1:                    # for ROM
                        fi = self.mem_w(org + 0x04)
                    else:                           # for binary
                        fi = self.mem_w(org + 0x1C)
                    while pc < fi:
                        c = 'DEF '
                        if self.mem_w(pc) in self._lsymbol:
                            c += self._lsymbol[self.mem_w(pc)]
                        elif self.mem_w(pc) in self._symbol:
                            c += self._symbol[self.mem_w(pc)]
                        else:
                            c += _adr(self.mem_w(pc))
                        h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                        _emit(pc, h, c, disp)
                        pc += 2
                else:                           # usual data
                    fi = _ncode(pc)                 # find address of following code section
                    t = 0                           # for block of DEF's (not used)
                    if pc in self._lsymbol:             # is a local sumbol existing ?
                        if self._lsymbol[pc][0] == 'X':     # an X() symbol ?
                            t = 1                               # a bunch of DEF ?
                        elif self._lsymbol[pc][0] == 'D':   # a D() symbol ? 
                            t = 1                               # one DEF?
                    while pc < fi:                  # loop over datas
                        ad = pc
                        done = 0
                        r = ''
                        if fi - pc >= 2:                # block larger then 2 bytes
                            # 1 word inside original binary and != 0x0000 -> DEF ? 
                            if (self.mem_w(pc) >= org_0) and (self.mem_w(pc) < end_0) and (self.mem_w(pc) != 0x0000):
                                if self.mem_w(pc) in self._lsymbol:     # is a local symbol existing ?
                                    c = 'DEF ' + self._lsymbol[self.mem_w(pc)]
                                    h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                                else:                                   # no 2 bytes
                                    c = 'BYT '  + _byt(self.mem_b(pc)) + ',' + _byt(self.mem_b(pc + 1))
                                    h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                                    r = 'R' + _adr(self.mem_w(pc))      # comment
                                    _addpsym(r, self.mem_w(pc))         # add a possible R(eference) symbol
                                pc += 2
                                done = 1                # partially done -> emit and loop again
                        if done == 0:               # not already done
                            if fi - pc == 1:            # one byte
                                c = 'BYT ' + _byt(self.mem_b(pc))
                                h = _byt(self.mem_b(pc)) + ' '
                                pc += 1
                            elif fi - pc == 2:          # one word
                                c = 'BYT ' + _byt(self.mem_b(pc)) + ',' + _byt(self.mem_b(pc + 1))
                                h = _byt(self.mem_b(pc)) + ' ' + _byt(self.mem_b(pc + 1)) + ' '
                                pc += 2
                            elif _bsz(pc) > pc:         # BSZ ?
                                c = 'BSZ ' + _adr(_bsz(pc) - pc)
                                h = ''
                                fi = _bsz(pc)
                                while pc < fi:
                                    h += _byt(self.mem_b(pc)) + ' '
                                    pc += 1
                            elif _asp(pc) > pc:         # ASP ?
                                c = 'ASP "' + _str(pc, _asp(pc)) +'"'
                                h = ''
                                fi = _asp(pc)
                                while pc < fi:
                                    h += _byt(self.mem_b(pc)) + ' '
                                    pc += 1
                            elif _asc(pc) > pc:         # ASC ?
                                c = 'ASC "' + _str(pc, _asc(pc)) +'"'
                                h = ''
                                fi = _asc(pc)
                                while pc < fi:
                                    h += _byt(self.mem_b(pc)) + ' '
                                    pc += 1
                            else:                       # just a bunch of bytes
                                c = 'BYT ' + _byt(self.mem_b(pc))
                                h = _byt(self.mem_b(pc)) + ' '
                                pc += 1
                                while (pc not in self._ldad) and (chr(self.mem_b(pc)) not in self._asc) and (pc < fi):
                                    c += ',' + _byt(self.mem_b(pc))
                                    h += _byt(self.mem_b(pc)) + ' '
                                    pc += 1
                        _emit(ad, h, c, disp, r)    # emit opcode
            if (level >= 1) and (pc == end):    # loop again ?
                break
        # end of de-assembly
        if (level >= 1):    # emit FIN opcode
            _emit(pc, '', 'FIN', disp)
        return self._psymbol    # return possible local symbol
    def ass_file(self, file_in, file_out=None, base=8, disp=True):
        # file_in  : name of text file to assemble
        # file_out : name of binary file to save in or None
        # base     : base in which produce the assember listing
        # disp     : to produce the assembler listing
        #
        # always create relocatable binary
        # the file is created with a 0x200 bytes header to ease the use of LIF utils
        #
        # auxilliary function using line return : something, remaining_line
        #
        def _get_chr(line):             # get a byte from line
            return ord(line[0]), line[1:]
        def _get_adr(line):             # get an address from line in base (constant)
            i = 0
            adr = 0
            ba = base
            if ba == 8:                             # base 8
                j = 0
                while line[j] in '0123456789ABCDEF':    # one token
                    j += 1
                if line[j-1] == 'D':                # last char D ?
                    ba = 10                             # base 10
                    j -= 1
                elif line[j-1] == 'C':              # last char C ?
                    ba = 11                             # base 16
                    j -= 1
            if ba == 8:
                while line[i] in '01234567':
                   adr = (adr * 8) + eval('0' + line[i])
                   i += 1
            elif ba == 10:      # special case for base 10
                while i < j:
                   adr = (adr * 10) + eval(line[i])
                   i += 1
                i += 1
            elif ba == 11:      # special case for base 16
                while i < j:
                   adr = (adr * 16) + eval('0x'+line[i])
                   i += 1
                i += 1
            elif ba == 16:
                while line[i] in '01234567890ABCDEF':
                   adr = (adr * 16) + eval('0x'+line[i])
                   i += 1
            else:
                raise AsmError, 'BASE ERROR'
            if adr > 0xFFFF:
                raise AsmError, 'TOO BIG'
            if line[i] not in (' ', '\t', ',','\n'):
                raise AsmError, 'SYNTAX'
            return adr, string.lstrip(line[i:])
        def _get_lab_adr(line):         # get an address from line (symbol or constant)
            i = 0
            while line[i] not in  (' ', '\t', ',', '\n'):
                i += 1
            if line[:i] in self._lldad:
                return self._lldad[line[:i]], string.lstrip(line[i:])
            elif line[:i] in self._ldad:
                if line[:i] not in self._lref:
                    self._lref[line[:i]] = [de]
                else:
                    self._lref[line[:i]].append(de)
                return self._ldad[line[:i]], string.lstrip(line[i:])
            elif line[:i] in self._dad:
                return self._dad[line[:i]], string.lstrip(line[i:])
            else:
                ad, line = _get_adr(line)
                return ad, string.lstrip(line)
        def _get_sep(line):             # get a separator (',') from line
            if len(line) > 0:
                if line[0] == ',':
                    return True, string.lstrip(line[1:])
            return False, line
        def _get_quot(line):            # get a double quote '"' from line
            i = 0
            while (line[i] != '"') and i < len(i):
                i += 1
            if line[i] == '"':
                return True, line[i + 1:]
            return False, line
        def _get_bytes(line):           # get a list of bytes from line
            byt, line = _get_lab_adr(line)
            bytes = [byt]
            sep, line = _get_sep(line)
            while sep:
                byt, line = _get_lab_adr(line)
                bytes.append(byt)
                sep, line = _get_sep(line)
            return bytes, string.lstrip(line)
        def _count_lab_adr(line):       # get an address from line (symbol or constant) in first pass (the symbol isn't necessary defined)
            i = 0
            while line[i] not in  (' ', '\t', ',', '\n'):
                i += 1
            if line[:i] in self._lldad:
                return self._lldad[line[:i]], string.lstrip(line[i:])
            elif line[:i] in self._ldad:
                return self._ldad[line[:i]], string.lstrip(line[i:])
            elif line[:i] in self._dad:
                return self._dad[line[:i]], string.lstrip(line[i:])
            else:
                try:
                    ad, line = _get_adr(line)
                except AsmError, detail:
                    if detail.value != 'SYNTAX':
                        raise
                    else:
                        print 'SYMBOL NOT DEFINED'
                        ad = 0xFFFF
                        line = string.lstrip(line[i:])
                return ad, string.lstrip(line)
        def _count_bytes(line):         # get a list of bytes from line in first pass
            byt, line = _count_lab_adr(line)
            bytes = [byt]
            sep, line = _get_sep(line)
            while sep:
                byt, line = _count_lab_adr(line)
                bytes.append(byt)
                sep, line = _get_sep(line)
            return bytes, string.lstrip(line)
        def _get_rp(line, pref='R'):    # get an RP register (Roo or Xoo) from line
            if line[0] == pref:
                if line[1] == '*':
                    r = 1
                    i = 2
                elif line[1] == '#':
                    r = -1
                    i = 2
                else:
                    i = 1
                    while line[i] in '012345678':
                        i += 1
                    r = eval('0'+line[1:i])
                    if (r == 0x1) or (r >= 0x40):
                        raise AsmError, 'RP FALSE'
                if line[i] not in (' ', '\t', ',', '\n'):
                    raise AsmError, 'RP UNKNOWN'
                return r, string.lstrip(line[i:])
            else:
                raise AsmError, 'NO RP'
        def _get_mode(opc):             # return the mode (B:0 or M:1) from opcode
            if opc[2] =='B':
                return 0
            elif opc[2] == 'M':
                return 1
            else:
                raise AsmError, 'UNKNOWN MODE'
        def _parse(line):               # parse a line (line #, lab, op, remaining_line)
            lab = ''                        # label
            op = ''                         # opcode
            i = 0
            while line[i] in '01234566789': # find a line # ?
                i += 1
            if i > 0:                           # found
                i = string.find(line, ' ')
                number = eval(line[:i])
                line = line[i + 1:]
            else:                               # not found
                number = -1
            i = 0
            if line[0] != '!':              # not a comment ?
                if line[0] != ' ':              # is there a label ?
                    i = 0
                    while line[i] not in (' ', '\t', '\n'):     # label
                        i += 1
                    lab = line[:i]
                    line = string.lstrip(line[i:])
                    i = 0
                    while line[i] not in (' ', '\t', '\n'):     # opcode
                        i += 1
                    op = line[:i]
                    line = string.lstrip(line[i:])
                else:                           # no label
                    line = string.lstrip(line)
                    i = 0
                    while line[i] not in  (' ', '\t', '\n'):    # opcode
                        i += 1
                    op = line[:i]
                    line = string.lstrip(line[i:])
            return number, lab, op, line
        def _out(code):                 # write to file_out if needed
            if file_out != None:
                fo.write(chr(code))
        def _adr(ad):                   # convert an address in base
            if base == 8:
                d = '00000' + oct(ad)
                return d[-6:]
            else:
                return '%04X' % ad
        def _emit(ad, code, lab, op, mline, disp):  # emit the line for assembler listing
            h = ''              # string of code in base
            for c in code:      # create it in base
                _out(c)
                if base == 8:
                    d = '00' + oct(c) 
                    h += d[-3:] + ' '
                else:
                    h += '%02X ' % c
            if disp:            # listing ?
                first = 1           # first line
                while (h != '') or (first == 1):
                    if first == 1:      # first line with label and comment
                        if base == 8:
                            print _adr(ad) + ' ' + h[:8] + string.ljust('', 8-len(h[:8])) + string.ljust(lab,9) + op + ' ' + mline[:-1]
                        else:
                            print _adr(ad) + ' ' + h[:6] + string.ljust('', 6-len(h[:6])) + string.ljust(lab,9) + op + ' ' + mline[:-1]
                        first = 0
                    else:               # further lines
                        if base == 8:
                            print _adr(ad) + ' ' + h[:8] + string.ljust('', 8-len(h[:8]))
                        else:
                            print _adr(ad) + ' ' + h[:6] + string.ljust('', 6-len(h[:6]))
                    if base == 8:
                        h = h[8:]
                    else:
                        h = h[6:]
                    ad += 2
        fd = open(file_in, 'r')         # preliminary pass for EQU and DAD
        for line in fd:                 # for all lines
            n_line, lab, op, line = _parse(line)
            if op == 'EQU':                 # op == EQU ?
                if lab != '':
                    if lab in self._ldad:
                        raise AsmError, 'DUPLICATE SYMBOL'
                    else:
                        ad, line = _get_adr(line)
                        self._ldad[lab] = ad
                else:
                    raise AsmError, 'NO SYMBOL'
            elif op == 'DAD':               # op == DAD ?
                if lab != '':
                    if lab not in self._dad:
                        if lab in self._ldad:
                            raise AsmError, 'DUPLICATE SYMBOL'
                        else:
                            ad, line = _get_adr(line)
                            self._ldad[lab] = ad
                else:
                    raise AsmError, 'NO SYMBOL'
        fd.close()
        level = 0                       # first pass
        while level != 2:               # loop until second pass done
            print 'PASS', level
            bin_begin = 0               # start address of binary
            bin_end = -1                # end address of binary
            if (level == 1) and (file_out != None): # header ? (second pass with output)
                fo = open(file_out, 'wb')
                l_h = self.lif_header(bin_size, type ='BPGM')
                for code in l_h:
                    fo.write(chr(code))
            fd = open(file_in, 'r')
            done = 0
            org = 0x0000
            pc = org
            i_line = 1      # # of lines
            lldad = {}      # store '.' local labels in first pass
            mlldad = ''     # store local label for the scope of '.' labels
            for line in fd: # for all lines
                h = []          # list of bytes to emit
                de = pc         # start address of current line
                n_line, lab, op, line = _parse(line)    # parse line
                if n_line == -1:                        # is there a number ?
                    n_line = i_line                         # no, create one
                    i_line += 1
                mline = line
                if op != '':    # is there an opcode ?
                    try:            # to catch AsmError and display some diagnosis
                        if op == 'EQU':                 # already done, just skip stuff in line
                            ad, line = _get_adr(line)
                        elif op == 'DAD':               # already done, just skip stuff in line
                            ad, line = _get_adr(line)
                        else:                           # standard label
                            if lab != '':
                                if level == 0:
                                    if lab in self._ldad:
                                        raise AsmError, 'DUPLICATE LABEL'
                                    else:
                                        if lab[0] == '.':       # local label
                                            if lab in lldad:
                                                raise AsmError, 'DUPLICATE LOCAL LABEL'
                                            lldad[lab] = pc
                                        else:
                                            if lldad != {}:
                                                self._llsymbols[mlldad] = lldad
                                            lldad = {}
                                            mlldad = lab
                                            self._ldad[lab] = pc
                                else:
                                    if lab in self._llsymbols:
                                        self._lldad = self._llsymbols[lab]
                                    if lab in self._lldad:
                                        if self._lldad[lab] != pc:
                                            raise AsmError, 'MOVING LABEL [%04X -> %04X]' % (self._ldad[lab], pc,)
                                    elif self._ldad[lab] != pc:
                                            raise AsmError, 'MOVING LABEL [%04X -> %04X]' % (self._ldad[lab], pc,)
                        if op == 'ORG':
                            pc, line = _get_adr(line)
                            bin_begin = pc
                        elif op == 'FIN':               # end of assembly
                            done = 1
                            bin_end = de
                            break
                        elif op == 'DEF':
                            if level == 1:
                                ad, line = _get_lab_adr(line)
                                h.append(ad % 0x100)
                                h.append(ad / 0x100)
                            pc += 2
                        elif op == 'BYT':
                            if level == 1:              # get list of bytes all defined (second pass)
                                bytes, line = _get_bytes(line)
                            else:                       # get list of bytes (some can be undefined as symbol, first pass)
                                bytes, line = _count_bytes(line)
                            for b in bytes:
                                if b > 0xFF:
                                    raise AsmError, 'TOO BIG'
                                if level == 1:
                                    h.append(b)
                                pc += 1
                        elif op == 'VAL':
                            if level == 1:
                                ad, line = _get_lab_adr(line)
                            else:
                                ad, line = _count_lab_adr(line)
                            if ad > 0xFF:                   # case of word
                                if level == 1:
                                    h.append(ad % 0x100)
                                    h.append(ad / 0x100)
                                pc += 2
                            else:                           # case of byte
                                if level == 1:
                                    h.append(ad)
                                pc += 1
                        elif op == 'NAM':               # Header of binary ....
                            if level == 1:                  # second pass, do it really
                                ad, line = _get_adr(line)       # get code
                                if ad > 0xFF:
                                    raise AsmError, 'TOO BIG'
                                ok, line = _get_sep(line)
                                if not ok:
                                    raise AsmError, 'SYNTAX'
                                i = 0
                                while (line[i] not in (' ', '\t', '\n')) and (i < 4):   # get name
                                    h.append(ord(line[i]))
                                    pc += 1
                                    i += 1
                                line = string.lstrip(line[i:])
                                while i < 4:
                                    h.append(0x00)
                                    pc += 1
                                    i += 1
                                h.append(bin_size % 0x100)      # append binary size
                                h.append(bin_size / 0x100)
                                pc += 2
                                i += 2
                                h.append(0x02)                  # append 0x02 .... need informations about it
                                h.append(ad)                    # append code
                                pc += 2 
                                i += 2
                                while i < 18:                   # append file name (only ' ')
                                    h.append(0x20)
                                    pc += 1
                                    i += 1
                                while i < 24:                   # fill until 24 bytes
                                    h.append(0x00)
                                    pc += 1
                                    i += 1
                                h.append(org % 0x100)           # append org of binary
                                h.append(org / 0x100)
                                pc += 2
                                i += 2
                            else:                           # first pass, only count its size 
                                pc += 26
                        elif op == 'ASC':
                            ok, line = _get_quot(line)
                            if ok:
                                i = 0
                                while (line[i] != '"') and (i < len(line)):
                                    if level == 1:
                                        h.append(ord(line[i]))
                                    i += 1
                                if line[i] != '"':
                                    raise AsmError, 'NO ENDING "'
                                pc += i
                                line = string.lstrip(line[i + 1:])
                            else:
                                raise AsmError, 'NO FIRST "'
                        elif op == 'ASP':
                            ok, line = _get_quot(line)
                            if ok:
                                i = 0
                                while (line[i] != '"') and (i < len(line)):
                                    if level == 1:
                                        h.append(ord(line[i]))
                                    i += 1
                                if line[i] != '"':
                                    raise AsmError, 'NO ENDING "'
                                if level == 1:
                                    h[-1] += 0x80
                                pc += i
                                line = string.lstrip(line[i + 1:])
                            else:
                                raise AsmError, 'NO FIRST "'
                        elif op == 'BSZ':
                            nb, line = _get_adr(line)
                            if level == 1:
                                for i in range(nb):
                                    h.append(0x00)
                            pc += nb
                        elif op == 'GTO':
                            if self.drp() != 4:             # current DRP isn't 4 ?
                                if level == 1:                  # do it in second pass
                                    h.append(0x44)                  # ADM R4,=
                                self.drp(4)                     # adjust DRP
                                pc += 1                         # one byte
                            if level == 1:                  # do it in second pass
                                ad, line = _get_lab_adr(line)
                                of = ad - (pc + 3)
                                if of < 0:
                                    of = 0x10000 + of
                                h.append(0xCB)
                                h.append(of % 0x100)
                                h.append(of / 0x100)
                            pc += 3
                            self.arp(-1)                    # for next line, adjust DRP & ARP as unknown
                            self.drp(-1)
                        elif op == 'ARP':
                            r, line = _get_rp(line)
                            if level == 1:
                                if r >= 0x0:
                                    h.append(r)
                                else:
                                    raise AsmError, 'R# NOT ALLOWED'
                            self.arp(r)
                            pc += 1
                        elif op == 'DRP':
                            r, line = _get_rp(line)
                            if level == 1:
                                if r >= 0x0:
                                    h.append(0x40 + r)
                                else:
                                    raise AsmError, 'R# NOT ALLOWED'
                            self.drp(r)
                            pc += 1
                        elif op in ('BIN', 'BCD', 'SAD', 'DCE', 'ICE', 'CLE', 'RTN', 'PAD'):
                            i = ['BIN', 'BCD', 'SAD', 'DCE', 'ICE', 'CLE', 'RTN', 'PAD'].index(op)
                            if (i == 6) or (i == 7):    # for RTN and PAP
                                self.drp(-1)                # adjust ARP & DRP as unknown
                                self.arp(-1)
                            if level == 1:
                                h.append(0x98 + i)
                            pc += 1
                        elif op[:2] in ('EL', 'ER', 'LL', 'LR', 'IC', 'DC', 'TC', 'NC', 'TS', 'CL'):
                            if len(op) > 3:
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.drp()) and (r != -1):
                                if level == 1:
                                    h.append(0x40 + r)
                                self.drp(r)
                                pc += 1
                            if level == 1:
                                i = ['EL', 'ER', 'LL', 'LR', 'IC', 'DC', 'TC', 'NC', 'TS', 'CL'].index(op[:2])
                                h.append(0x80 + i * 2 + _get_mode(op))
                            pc += 1
                        elif op[:2] in ('OR', 'XR'):
                            if len(op) > 3:
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.drp()) and (r != -1):
                                if level == 1:
                                    h.append(0x40 + r)
                                self.drp(r)
                                pc += 1
                            ok, line = _get_sep(line)
                            if not ok:
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.arp()) and (r != -1):
                                if level == 1:
                                    h.append(0x00 + r)
                                self.arp(r)
                                pc += 1
                            if level == 1:
                                i = ['OR', 'XR'].index(op[:2])
                                h.append(0x94 + i * 2 + _get_mode(op))
                            pc += 1
                        elif op[:2] in ('LD', 'ST'):
                            m = _get_mode(op)
                            i = ['LD', 'ST'].index(op[:2])
                            d = 0
                            if len(op) == 4:
                                if op[-1] == 'I':
                                    d = 2
                                elif op[-1] == 'D':
                                    d = 1
                                else:
                                    raise AsmError, 'SYNTAX'
                            elif len(op) != 3:
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.drp()) and (r != -1):
                                if level == 1:
                                    h.append(0x40 + r)
                                self.drp(r)
                                pc += 1
                            ok, line = _get_sep(line)
                            if not ok:
                                raise AsmError, 'SYNTAX'
                            if line[0] == '=':                          # ST/LDmd R,=
                                ch, line = _get_chr(line)
                                if level == 1:
                                    h.append(0xA8 + i * 2 + m + d * 8)
                                pc += 1
                                if (m == 0) and (d == 0) and (i == 0):      # LDB R,=
                                    n = 1
                                elif (d == 0) and (i == 1):                 # STx R,=
                                    raise AsmError, 'SYNTAX : no semantic'
                                elif (m == 1) and (d == 0) and (i == 0):    # LDM R,=
                                    if (self.drp() < 0) or (self.drp() == 1):
                                        raise AsmError, 'CAN T COMPUTE DRP'
                                    elif self.drp() < 0x20:
                                        n = 2 - (self.drp() % 2)
                                    else:
                                        n = 8 - (self.drp() % 8)
                                else:
                                    n = 2
                                if n == 1:
                                    if level == 1:
                                        ad, line = _get_lab_adr(line)
                                        if ad > 0xFF:
                                            raise AsmError, 'TOO BIG'
                                        h.append(ad)
                                    n -= 1
                                    pc += 1
                                else:
                                    if level == 1:
                                        bytes, line = _get_bytes(line)
                                    else:
                                        bytes, line = _count_bytes(line)
                                    for b in bytes:
                                        if (b > 0xFF) or ((n == 2) and (len(bytes) == 1)):
                                            if level == 1:
                                                h.append(b % 0x100)
                                                h.append(b / 0x100)
                                            n -= 2
                                            pc += 2
                                        else:
                                            if level == 1:
                                                h.append(b)
                                            n -= 1
                                            pc += 1
                                if n < 0:
                                    raise AsmError, 'TOO MUCH DATA (PC=%04X)' % pc
                                if n > 0:
                                    print 'NOT ENOUGH DATA (%d)' % n
                            elif line[0] == 'X':                            # ST/LDmD/I R,X,a
                                r, line = _get_rp(line, 'X')
                                if d == 0:
                                    raise AsmError, 'MODE NOT ALLOWED'
                                if (r != self.arp()) and (r != -1):
                                    if level == 1:
                                        h.append(0x00 + r)
                                    self.arp(r)
                                    pc += 1
                                ok, line = _get_sep(line)
                                if not ok:
                                    raise AsmError, 'SYNTAX'
                                if level == 1:
                                    ad, line = _get_lab_adr(line)
                                    h.append(0xAC + i * 2 + m + d * 8)
                                    h.append(ad % 0x100)
                                    h.append(ad / 0x100)
                                pc += 3
                            elif line[0] == 'R':                            # ST/LDmd R,R
                                r, line = _get_rp(line)
                                if (r != self.arp()) and (r != -1):
                                    if level == 1:
                                        h.append(0x00 + r)
                                    self.arp(r)
                                    pc += 1
                                if level == 1:
                                    if d == 2:
                                        d = 3
                                    h.append(0xA0 + i * 2 + m + d * 4)
                                pc += 1
                            else:
                                raise AsmError, 'SYNTAX'
                            if (self.drp() == 4) and (i == 0):
                                self.drp(-1)
                                self.arp(-1)
                        elif op[:2] in ('CM', 'AD', 'SB', 'AN'):
                            m = _get_mode(op)
                            i = ['CM', 'AD', 'SB', 'AN'].index(op[:2])
                            d = 0
                            if len(op) == 4:
                                if op[-1] == 'D':
                                    d = 1
                                else:
                                    raise AsmError, 'SYNTAX'
                            elif len(op) != 3:
                                raise AsmError, 'SYNTAX'
                            if (i == 3) and (m == 0):       # ANB not defined
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.drp()) and (r != -1):
                                if level == 1:
                                    h.append(0x40 + r)
                                self.drp(r)
                                pc += 1
                            ok, line = _get_sep(line)
                            if not ok:
                                raise AsmError, 'SYNTAX'
                            if line[0] == '=':              # immediate mode
                                ch, line = _get_chr(line)
                                if level == 1:
                                    h.append(0xC8 + i * 2 + m + d * 8)
                                pc += 1
                                if d == 1:      # compute number of bytes to follow (n)
                                    n = 2
                                elif m == 0:
                                    n = 1
                                else:
                                    if (self.drp() < 0) or (self.drp() == 1):
                                        raise AsmError, 'CAN T COMPUTE DRP'
                                    elif self.drp() < 0x20:
                                        n = 2 - (self.drp() % 2)
                                    else:
                                        n = 8 - (self.drp() % 8)
                                if n == 1:
                                    if level == 1:
                                        ad, line = _get_lab_adr(line)
                                        if ad > 0xFF:
                                            raise AsmError, 'TOO BIG'
                                        h.append(ad)
                                    n -= 1
                                    pc += 1
                                elif (n == 2) and (d == 1):
                                    if level == 1:
                                        ad, line = _get_lab_adr(line)
                                        h.append(ad % 0x100)
                                        h.append(ad / 0x100)
                                    n -= 2
                                    pc += 2
                                else:
                                    if level == 1:
                                        bytes, line = _get_bytes(line)
                                    else:
                                        bytes, line = _count_bytes(line)
                                    for b in bytes:
                                        if (b > 0xFF) or ((n == 2) and (len(bytes) == 1)):
                                            if level == 1:
                                                h.append(b % 0x100)
                                                h.append(b / 0x100)
                                            n -= 2
                                            pc += 2
                                        else:
                                            if level == 1:
                                                h.append(b)
                                            n -= 1
                                            pc += 1
                                if n < 0:
                                    raise AsmError, 'TOO MUCH DATA'
                                if n > 0:
                                    print 'NOT ENOUGH DATA (%d)' % n
                            elif line[0] == 'R':            # register mode
                                r, line = _get_rp(line)
                                if (r != self.arp()) and (r != -1):
                                    if level == 1:
                                        h.append(0x00 + r)
                                    self.arp(r)
                                    pc += 1
                                if level == 1:
                                    h.append(0xC0 + i * 2 + m + d * 24)
                                pc += 1
                            else:
                                raise AsmError, 'SYNTAX'
                            if (self.drp() == 4) and (i > 0):   # changing R4 ? (GOTO)
                                self.drp(-1)                        # adjusr ARP & DRP as unknown 
                                self.arp(-1)
                        elif op == 'JSB':
                            if line[0] == '=':          # direct mode
                                ch, line = _get_chr(line)
                                if level == 1:
                                    ad, line = _get_lab_adr(line)
                                    h.append(0xCE)
                                    h.append(ad % 0x100)
                                    h.append(ad / 0x100)
                                self.drp(-1)
                                self.arp(-1)
                                pc += 3
                            elif line[0] == 'X':        # indexed mode
                                r, line = _get_rp(line, 'X')
                                if (r != self.arp()) and (r != -1):
                                    if level == 1:
                                        h.append(0x00 + r)
                                    self.arp(r)
                                    pc += 1
                                ok, line = _get_sep(line)
                                if not ok:
                                    raise AsmError, 'SYNTAX'
                                if level == 1:
                                    ad, line = _get_lab_adr(line)
                                    h.append(0xC6)
                                    h.append(ad % 0x100)
                                    h.append(ad / 0x100)
                                self.drp(-1)
                                self.arp(-1)
                                pc += 3
                            else:
                                raise AsmError, 'SYNTAX'
                        elif op[:2] in ('PO', 'PU'):
                            m = _get_mode(op)
                            i = ['PO', 'PU'].index(op[:2])
                            if len(op) == 4:
                                if op[-1] == 'I':
                                    d = 1
                                elif op[-1] == 'D':
                                    d = 0
                                else:
                                    raise AsmError, 'SYNTAX'
                            else:
                                raise AsmError, 'SYNTAX'
                            r, line = _get_rp(line)
                            if (r != self.drp()) and (r != -1):
                                if level == 1:
                                    h.append(0x40 + r)
                                self.drp(r)
                                pc += 1
                            ok, line = _get_sep(line)
                            if not ok:
                                raise AsmError, 'SYNTAX'
                            if line[0] == '+':
                                s = 0
                            elif line[0] == '-':
                                s = 1
                            else:
                                raise AsmError, 'SYNTAX'
                            ch, line = _get_chr(line)
                            r, line = _get_rp(line)
                            if (r != self.arp()) and (r != -1):
                                if level == 1:
                                    h.append(0x00 + r)
                                self.arp(r)
                                pc += 1
                            if level == 1:
                                h.append(0xE0 + s * 2 + m + i * 4 + d * 8)
                            pc += 1
                        elif op[:1] == 'J':
                            if op[1:] in ('MP', 'NO', 'OD', 'EV', 'NG', 'PS', 'NZ', 'ZR', 'EN', 'EZ', 'NC', 'CY', 'LZ', 'LN', 'RZ', 'RN'):
                                if level == 1:
                                    i = ['MP', 'NO', 'OD', 'EV', 'NG', 'PS', 'NZ', 'ZR', 'EN', 'EZ', 'NC', 'CY', 'LZ', 'LN', 'RZ', 'RN'].index(op[1:])
                                    h.append(0xF0 + i)
                                    adr, line = _get_lab_adr(line)
                                    of = adr - (pc + 2)
                                    if (of > 127) or (of < -128):
                                        raise AsmError, 'JMP TOO LONG'
                                    if of < 0:
                                        of = 0x100 + of
                                    h.append(of)
                                if op[1:] == 'MP':
                                    self.arp(-1)
                                    self.drp(-1)
                                pc += 2
                        elif op == 'EQU':   # already done
                            pass
                        elif op == 'DAD':   # already done
                            pass
                        else:       # unknow opcode
                            raise AsmError, 'SYNTAX'
                        if (level > 0) and (len(line) > 0):     # some consistency checks
                            if (line[0] != '!') and (line[0] != '\n'):
                                print 'WARNING : garbage in line %d [%s]\n' % (n_line, line,)
                        if level == 1:  # emit line in second pass
                            _emit(de, h, lab, op, mline, disp)
                    except: # error treatment
                        print n_line, lab, op, mline
                        raise
                    
            fd.close()      # finished for all lines
            if level == 0:  # first pass ?
                bin_size = bin_end - bin_begin  # binary size
            else:
                if (bin_end - bin_begin) != bin_size:   # consistency check (first and second pass)
                    raise AsmError, 'BINARY SIZE CHANGED'
            if (level == 1) and (file_out != None):     # padding at the end of the file (256 bytes boundary)
                i = bin_size
                print '%04X' % i
                while (i % 0x100) != 0:
                    fo.write(chr(0))
                    i += 1
                print '%04X' % i
                fo.flush()
                fo.close()
            level += 1  # next pass
        print 'Done'
    def lif_header(self, size, type ='BPGM'):
        # create the list of bytes for the lif_header
        def _append_b(c):               # append a byte
            for b in c:
                h.append(b)
        def _append_c(c):               # append a string
            for i in range(len(c)):
                h.append(ord(c[i]))
        def _append_z(n, code= 0x00):   # append a block of 'code'
            for i in range(n):
                h.append(code)
        def _append_w(w):               # append a word
            h.append(w / 0x100)
            h.append(w % 0x100)
        def _append_3(w):               # append a 3 bytes quantity
            h.append(w % 0x100)
            h.append((w / 0x100) % 0x100)
            h.append(w / 0x10000)
        h = []                      # the header
        n_c = (size + 0xFF) / 0x100
        _append_b((0x80, 0x00,))
        _append_c('HFSLIF')
        _append_b((0x00, 0x00, 0x00, 0x01, 0x10, 0x00, 0x00, 0x00))
        _append_b((0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01))
        _append_z(2)
        _append_w(n_c + 2)
        _append_z(6, 0x11)
        _append_z(0xCE)
        _append_z(6, 0x11)
        _append_z(2)
        _append_c('WS_FILE   ')
        if type == 'BPGM':
            _append_w(0xE008)
        elif type == 'PROG':
            _append_w(0xE020)
        elif type == 'DATA':
            _append_w(0xE010)
        else:
            raise AsmError, 'TYPE UNKNOWN'
        _append_b((0x00, 0x00, 0x00, 0x02))
        _append_z(2)
        _append_w(n_c + 1)
        _append_z(6, 0x99)
        _append_w(0x8001)
        _append_3(size)
        _append_b((0x01,))
        _append_z(10)
        _append_z(2, 0xFF)
        _append_z(0xD4)
        return h
